/*-----------------------------------------------------------------------
| pcibios.c - interface to pci-bios function calls.
|  pci bios calls can be real-mode, 16-bit protected, 32bit protected.
|------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <dos.h>

/********************* base types *************************/
#ifndef  NO_BASE_PCI_TYPES
/* Miscellaneous */
#define FALSE   0
#define TRUE    1

/* Definitions so that BYTE is 8-bits, WORD is 16-bits, and DWORD is 32-bits*/
typedef unsigned char   BYTE ;
typedef unsigned short  WORD ;
typedef unsigned long   DWORD ;
typedef union x86Regs   X86REG ;

/* x86 Registers */
struct DWordRegs  {
  DWORD   eax ;
  DWORD   ebx ;
  DWORD   ecx ;
  DWORD   edx ;
  DWORD   esi ;
  DWORD   edi ;
  } ;

struct WordRegs  {
  WORD    ax ;
  WORD    empty1 ;
  WORD    bx ;
  WORD    empty2 ;
  WORD    cx ;
  WORD    empty3 ;
  WORD    dx ;
  WORD    empty4 ;
  WORD    si ;
  WORD    empty5 ;
  WORD    di ;
  WORD    empty6 ;
  WORD    cflag ;
  };

struct ByteRegs  {
  BYTE    al, ah ;
  WORD    empty1 ;
  BYTE    bl, bh ;
  WORD    empty2 ;
  BYTE    cl, ch ;
  WORD    empty3 ;
  BYTE    dl, dh ;
  };

union x86Regs  {
  struct DWordRegs        e ;
  struct WordRegs         x ;
  struct ByteRegs         h ;
  } ;
#endif

int pci_bios(X86REG *InRegs, X86REG *OutRegs);

/* int 0x1a sub-function identifier for PCI config functions */
#define PCI_FUNCTION_ID   0xb1

#include "pcibios.h"
#include "rcktldef.h"

int processor_mode = REAL_MODE;

/*-----------------------------------------------------------------------
| pci_init.c - initialize BIOS pci facility.
|------------------------------------------------------------------------*/
int pci_init(void)
{
 int stat;
 static int major_version, minor_version, last_bus, access_flags;

  stat = pci_get_status((int *) &major_version,
                        (int *) &minor_version,
                        (int *) &last_bus,
                        (int *) &access_flags);
  return stat;
}

/*-----------------------------------------------------------------------
 pci_bios - bios call for PCI related stuff.
|------------------------------------------------------------------------*/
int pci_bios(X86REG *InRegs, X86REG *OutRegs)
{
// int stat;
// union REGS in, out;
// struct SREGS sregs;
 WORD out_ax, out_bx, out_cx, out_dx, out_di, out_si, out_cflag;

  switch (processor_mode)
  {
    case REAL_MODE:
      out_cflag = 0;
      out_ax = InRegs->x.ax;
      out_bx = InRegs->x.bx;
      out_cx = InRegs->x.cx;
      out_dx = InRegs->x.dx;
      out_di = InRegs->x.di;
      out_si = InRegs->x.si;

      ASM {
        pusha
        push es
        push ds
        mov ax, out_ax;
        mov bx, out_bx;
        mov cx, out_cx;
        mov dx, out_dx;
        mov di, out_di;
        mov si, out_si;
        int 1AH  // call bios
        jnc no_car
        mov out_cflag, 1;
     no_car:
        mov out_ax, ax
        mov out_bx, bx
        mov out_cx, cx
        mov out_dx, dx
        pop ds
        pop es
        popa
      }
      OutRegs->e.eax = out_ax;
      OutRegs->e.ebx = out_bx;
      OutRegs->e.ecx = out_cx;
      OutRegs->e.edx = out_dx;
      OutRegs->x.cflag = out_cflag;

    return 0;  /* ok */

    case PROT16_MODE:
    case PROT32_MODE:
    default:
//      printf("Bios mode not implemented\n");
    return 1;  /* err */
  }
}

/*-----------------------------------------------------------------------
 pci_get_status -
|------------------------------------------------------------------------*/
int pci_get_status(int *major_version,
                   int *minor_version,
                   int *last_bus,
                   int *access_flags)
{
//  DWORD dstat;
 static X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID; /* 0xb1 */
  InRegs.h.al = 0x01;            /* PCI_BIOS_PRESENT */

  if (pci_bios((X86REG *) &InRegs,
               (X86REG *) &OutRegs) != 0) return 1;  /* err */

  if (OutRegs.h.ah != 0)
    return 1;  /* err */

//  if (OutRegs.x.cflag != 0)
//    return 1;  /* err */

/*  if (OutRegs.e.edx != 0x020494350)  // check for identifier: "PCI^" */
/* kludge in for 16-bit regs */
  if (OutRegs.e.edx != 0x4350)  /* check for identifier: "PCI^" */
  {
    return 2;
  }

  *major_version = OutRegs.h.bh;
  *minor_version = OutRegs.h.bl;
  *last_bus = OutRegs.h.cl;
  *access_flags = OutRegs.h.al;

  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_find_device -
|------------------------------------------------------------------------*/
int pci_find_device(int DevId,
                    int VenId,
                    int Index,
                    int *bus_number,
                    int *function_number,
                    int *device_number)
{
  static X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x02;             /* FIND_PCI_DEVICE */
  InRegs.x.cx = DevId;            /* device ID to find */
  InRegs.x.dx = VenId;            /* vendor ID to find */
  InRegs.x.si = Index;            /* index(next card..) */

  if (pci_bios((X86REG *) &InRegs,
               (X86REG *) &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }

  *bus_number = OutRegs.h.bh;
  *device_number = (OutRegs.h.bl >> 3);
  *function_number = (OutRegs.h.bl & 0x07);

  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_read_word - Read PCI config WORD value.
|------------------------------------------------------------------------*/
int pci_read_word(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  WORD *config_reg)
{
  static X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x09;             /* READ_CONFIG_WORD */
  InRegs.h.bh = BusNum ;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum ;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */

  if (pci_bios((X86REG *) &InRegs,
               (X86REG *) &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }

  *config_reg = OutRegs.x.cx;

  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_read_byte - Read PCI config BYTE value.
|------------------------------------------------------------------------*/
int pci_read_byte(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  BYTE *config_reg)
{
  static X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x08;             /* READ_CONFIG_BYTE */
  InRegs.h.bh = BusNum ;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum ;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */

  if (pci_bios((X86REG *) &InRegs,
               (X86REG *) &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }

  *config_reg = OutRegs.h.cl;

  return 0;  /* ok */
}

#ifdef UNUSED_FUNCS
/*-----------------------------------------------------------------------
 pci_read_dword - Read PCI config DWORD value.
|------------------------------------------------------------------------*/
int pci_read_dword(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  DWORD *config_reg)
{
  X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x0a;             /* READ_CONFIG_DWORD */
  InRegs.h.bh = BusNum ;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum ;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */

  if (pci_bios(&InRegs, &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }

  *config_reg = OutRegs.e.ecx;

  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_write_dword - Write PCI config DWORD value.
|------------------------------------------------------------------------*/
int pci_write_dword(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  DWORD config_reg)
{
  X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x0d;             /* WRITE_CONFIG_DWORD */
  InRegs.h.bh = BusNum;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */
  InRegs.e.ecx = config_reg;        /* data to write */

  if (pci_bios(&InRegs, &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }
  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_write_word - Write PCI config WORD value.
|------------------------------------------------------------------------*/
int pci_write_word(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  WORD config_reg)
{
  X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x0c;             /* WRITE_CONFIG_WORD */
  InRegs.h.bh = BusNum;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */
  InRegs.x.cx = config_reg;        /* data to write */

  if (pci_bios(&InRegs, &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }
  return 0;  /* ok */
}

/*-----------------------------------------------------------------------
 pci_write_byte - Write PCI config BYTE value.
|------------------------------------------------------------------------*/
int pci_write_byte(int BusNum,
                  int DevNum,
                  int FuncNum,
                  int RegNum,
                  BYTE config_reg)
{
  X86REG  InRegs, OutRegs;

  InRegs.h.ah = PCI_FUNCTION_ID ; /* 0xb1 */
  InRegs.h.al = 0x0b;             /* WRITE_CONFIG_BYTE */
  InRegs.h.bh = BusNum;
  InRegs.h.bl = ( DevNum << 3 ) | FuncNum;
  InRegs.x.di = RegNum;            /* index of register to read, 0,2,4.. */
  InRegs.h.cl = config_reg;        /* data to write */

  if (pci_bios(&InRegs, &OutRegs) != 0) return 1;  /* err */

  if(OutRegs.h.ah != 0)
  {
    return OutRegs.h.ah;  /* err */
  }
  return 0;  /* ok */
}
#endif
